#include <config.h>

#include "ftpd.h"
#include "ls_p.h"
#include "bsd-glob.h"
#include "messages.h"
#include "dynamic.h"
#include "ftpwho-update.h"
#include "globals.h"

#ifdef WITH_DMALLOC
# include <dmalloc.h>
#endif

static void wrstr(const int f, const char *s)
{
    static char outbuf[CONF_TCP_SO_SNDBUF];
    static size_t outcnt;
    size_t l;
    
    if (s == NULL) {
        if (outcnt > (size_t) 0U) {
            if (safe_write(f, outbuf, outcnt) != 0) {
                return;
            }
        }
        outcnt = (size_t) 0U;
        return;
    }
    if ((l = strlen(s)) <= (size_t) 0U) {
        return;
    }
    if (l <= (sizeof outbuf - outcnt)) {
        memcpy(outbuf + outcnt, s, l); /* secure, see above */
        outcnt += l;
        return;
    }
    if (outcnt < sizeof outbuf) {
        const size_t rest = sizeof outbuf - outcnt;
        
        memcpy(outbuf + outcnt, s, rest);   /* secure, see above */
        s += rest;
        l -= rest;
    }
    if (safe_write(f, outbuf, sizeof outbuf) != 0) {
        return;
    }
    while (l > sizeof outbuf) {
        if (safe_write(f, s, sizeof outbuf) != 0) {
            return;
        }
        s += sizeof outbuf;
        l -= sizeof outbuf;
    }
    if (l > (size_t) 0U) {
        memcpy(outbuf, s, l);          /* safe, l <= sizeof outbuf */
        outcnt = l;
    }
}

#ifdef NO_FTP_USERS
const char *getname(const uid_t uid)
{
    static char number[9];

    if (SNCHECK(snprintf(number, sizeof number, "%-8d", uid), sizeof number)) {
        _EXIT(EXIT_FAILURE);
    }
    return number;
}

const char *getgroup(const gid_t gid)
{
    static char number[9];

    if (SNCHECK(snprintf(number, sizeof number, "%-8d", gid), sizeof number)) {
        _EXIT(EXIT_FAILURE);
    }
    return number;
}
#else

const char *getname(const uid_t uid)
{
    struct userid *p;
    struct passwd *pwd = NULL;

    for (p = user_head; p; p = p->next) {
        if (p->uid == uid) {
            return p->name;
        }
    }
    if (chrooted == 0 && authresult.slow_tilde_expansion == 0) {
        pwd = getpwuid(uid);
    }
    if ((p = malloc(sizeof *p)) == NULL) {
        return "?       ";
    }
    p->uid = uid;
    p->name = malloc(9);

    if (p->name) {
        if (pwd) {
            if (SNCHECK(snprintf(p->name, 9, "%-8.8s", pwd->pw_name), 9)) {
                _EXIT(EXIT_FAILURE);
            }
        } else {
            if (SNCHECK(snprintf(p->name, 9, "%-8d", uid), 9)) {
                _EXIT(EXIT_FAILURE);
            }
        }
    }
    p->next = user_head;
    user_head = p;
    
    return p->name;
}

/* eeeehm... sorry for names, ya know copy&paste :))) */
const char *getgroup(const gid_t gid)
{
    struct groupid *p;
    struct group *pwd = NULL;

    for (p = group_head; p; p = p->next) {
        if (p->gid == gid) {
            return p->name;
        }
    }    
    if (chrooted == 0) {
        pwd = getgrgid(gid);
    }
    if ((p = malloc(sizeof *p)) == NULL) {
        return "?       ";        
    }
    p->gid = gid;
    p->name = malloc(9);
    if (p->name) {
        if (pwd) {
            if (SNCHECK(snprintf(p->name, 9, "%-8.8s", pwd->gr_name), 9)) {
                _EXIT(EXIT_FAILURE);
            }
        } else {
            if (SNCHECK(snprintf(p->name, 9, "%-8d", gid), 9)) {
                _EXIT(EXIT_FAILURE);
            }
        }
    }
    p->next = group_head;
    group_head = p;
    
    return p->name;
}
#endif

static void addfile(const char *name, const char *suffix)
{
    struct filename *p;
    int l;

    if (!name || !suffix) {
        return;
    }
    if (matches >= max_ls_files) {
        return;
    }
    matches++;
    l = (int) (strlen(name) + strlen(suffix));
    if (l > colwidth) {
        colwidth = l;
    }
    /* Fixed off-by-one -J. */
    if ((p = malloc(offsetof(struct filename, line) + l + 1U)) == NULL) {
        return;
    }
    if (SNCHECK(snprintf(p->line, l + 1U, "%s%s", name, suffix), l + 1U)) {
        _EXIT(EXIT_FAILURE);
    }
    if (tail) {
        tail->down = p;
    } else {
        head = p;
    }
    tail = p;
    filenames++;
}

/* listfile returns non-zero if the file is a directory */
static int listfile(const FileInfo * const fi,  const char *name)
{
    int rval = 0;
    struct stat st;
    struct tm *t;
    char suffix[2] = { 0, 0 };
    char m[MAXPATHLEN + 1U];    

#ifndef MINIMAL
    if (modern_listings != 0) {
        const char *n;
        char *alloca_nameline;
        const size_t sizeof_nameline = MAXPATHLEN + 256U;

        if (fi == NULL) {
            n = name;
        } else {
            n = FI_NAME(fi);
        }
        if ((alloca_nameline = ALLOCA(sizeof_nameline)) == NULL) {
            return 0;
        }
        if ((rval = modernformat(n, alloca_nameline, sizeof_nameline)) < 0) {
            ALLOCA_FREE(alloca_nameline);            
            return 0;
        }
        addfile(alloca_nameline, suffix);
        ALLOCA_FREE(alloca_nameline);
        
        return rval;
    }
#endif
    if (fi == NULL) {
        if (lstat(name, &st) < 0) {
            return 0;
        }    
    } else {
        st.st_size = fi->size;
        st.st_mtime = fi->mtime;        
        st.st_mode = fi->mode;
        st.st_nlink = fi->nlink;        
        st.st_uid = fi->uid;
        st.st_gid = fi->gid;
        name = FI_NAME(fi);
    }
#if defined(WITH_VIRTUAL_CHROOT) && defined(S_IFLNK) && defined(S_IFDIR)
    if (S_ISLNK(st.st_mode) && name[0] == '.' && name[1] == '.' && name[2] == 0) {
        st.st_mode &= ~S_IFLNK;
        st.st_mode |= S_IFDIR;
    }  /* Hack to please some Windows client that dislike ../ -> ../ */
#endif
    t = localtime((time_t *) & st.st_mtime);
    if (!t) {
        die(421, LOG_ERR, "localtime()");
    }
    if (opt_F) {
        if (S_ISLNK(st.st_mode))
            suffix[0] = '@';
        else if (S_ISDIR(st.st_mode)) {
            suffix[0] = '/';
            rval = 1;
        } else if (st.st_mode & 010101) {
            suffix[0] = '*';
        }
    }
    if (opt_l) {
        strncpy(m, " ---------", (sizeof m) - 1U);
        m[(sizeof m) - 1] = 0;
        switch (st.st_mode & S_IFMT) {
        case S_IFREG:
            m[0] = '-';
            break;
        case S_IFLNK:
            m[0] = 'l';
            break;            /* readlink() here? */
        case S_IFDIR:
            m[0] = 'd';
            rval = 1;
            break;
        }
        if (m[0] != ' ') {
            char *alloca_nameline;
            const size_t sizeof_nameline = MAXPATHLEN + MAXPATHLEN + 128U;
            char timeline[6U];
            
            if (st.st_mode & 0400) {
                m[1] = 'r';
            }
            if (st.st_mode & 0200) {
                m[2] = 'w';
            }
            if (st.st_mode & 0100) {
                m[3] = st.st_mode & 04000 ? 's' : 'x';
            } else if (st.st_mode & 04000) {
                m[3] = 'S';
            }
            if (st.st_mode & 040) {
                m[4] = 'r';
            }
            if (st.st_mode & 020) {
                m[5] = 'w';
            }
            if (st.st_mode & 010) {
                m[6] = st.st_mode & 02000 ? 's' : 'x';
            } else if (st.st_mode & 02000) {
                m[6] = 'S';
            }
            if (st.st_mode & 04) {
                m[7] = 'r';
            }
            if (st.st_mode & 02) {
                m[8] = 'w';
            }
            if (st.st_mode & 01) {
                m[9] = st.st_mode & 01000 ? 't' : 'x';
            } else if (st.st_mode & 01000) {
                m[9] = 'T';
            }            
            if (time(NULL) - st.st_mtime > 180 * 24 * 60 * 60) {
                if (SNCHECK(snprintf(timeline, sizeof timeline, "%5d",
                                     t->tm_year + 1900), sizeof timeline)) {
                    _EXIT(EXIT_FAILURE);
                }
            } else {
                if (SNCHECK(snprintf(timeline, sizeof timeline, "%02d:%02d",
                                     t->tm_hour, t->tm_min), sizeof timeline)) {
                    _EXIT(EXIT_FAILURE);
                }
            }
            if ((alloca_nameline = ALLOCA(sizeof_nameline)) == NULL) {
                return 0;
            }
#ifdef WITH_LARGE_FILES
            if (SNCHECK(snprintf(alloca_nameline, sizeof_nameline,
                                 "%s %4d %s %s %8llu %s %2d %s %s", 
                                 m, st.st_nlink, getname(st.st_uid),
                                 getgroup(st.st_gid), 
                                 (unsigned long long) st.st_size,
                                 months[t->tm_mon],
                                 t->tm_mday, timeline, name),
                        sizeof_nameline))
#else
            if (SNCHECK(snprintf(alloca_nameline, sizeof_nameline,
                                 "%s %4d %s %s %8lu %s %2d %s %s", 
                                 m, st.st_nlink, getname(st.st_uid),
                                 getgroup(st.st_gid), 
                                 (unsigned long) st.st_size,
                                 months[t->tm_mon],
                                 t->tm_mday, timeline, name),
                        sizeof_nameline))
#endif
            {
                ALLOCA_FREE(alloca_nameline);
                _EXIT(EXIT_FAILURE);
            }
            if (S_ISLNK(st.st_mode)) {
                char *p = alloca_nameline + strlen(alloca_nameline);
                
                {
                    int sx;
                    
                    if ((sx = readlink(name, m, sizeof m - 1U)) > 0) {
                        m[sx] = 0;
                    } else {
                        m[0] = m[1] = '.';
                        m[2] = 0;
                    }
                }
                suffix[0] = 0;
                if (opt_F && stat(name, &st) == 0) {
                    if (S_ISLNK(st.st_mode)) {
                        suffix[0] = '@';
                    } else if (S_ISDIR(st.st_mode)) {
                        suffix[0] = '/';
                    } else if (st.st_mode & 010101) {
                        suffix[0] = '*';
                    }
                }
                /* 2*MAXPATHLEN+gap should be enough, but be paranoid... */
                if (SNCHECK
                    (snprintf(p, (sizeof_nameline) - strlen(alloca_nameline),
                              " -> %s", m), 
                     (sizeof_nameline) - strlen(alloca_nameline))) {
                    ALLOCA_FREE(alloca_nameline);                    
                    _EXIT(EXIT_FAILURE);
                }
            }
            addfile(alloca_nameline, suffix);
            ALLOCA_FREE(alloca_nameline);            
        }                    /* hide non-downloadable files */
    } else {        
        if (S_ISREG(st.st_mode) ||
            S_ISDIR(st.st_mode) || S_ISLNK(st.st_mode)) {
            addfile(name, suffix);
        }
    }
    return rval;
}

static void outputfiles(int f)
{
    int n;
    struct filename *p;
    struct filename *q;

    if (!head) {
        return;
    }
    tail->down = NULL;
    tail = NULL;
    colwidth = (colwidth | 7) + 1;
    if (opt_l != 0 || opt_C == 0) {
        colwidth = 75;
    }
    /* set up first column */
    p = head;
    p->top = 1;
    if (colwidth > 75) {
        n = filenames;
    } else {
        n = (filenames + (75 / colwidth) - 1) / (75 / colwidth);
    }
    while (n && p) {
        p = p->down;
        if (p != NULL) {
            p->top = 0;
        }
        n--;
    }

    /* while there's a neighbour to the right, point at it */
    q = head;
    while (p) {
        p->top = q->top;
        q->right = p;
        q = q->down;
        p = p->down;
    }

    /* some are at the right end */
    while (q) {
        q->right = NULL;
        q = q->down;
    }

    /* don't want wraparound, do we? */
    p = head;
    while (p && p->down && !p->down->top) {
        p = p->down;
    }
    if (p && p->down) {
        p->down = NULL;
    }

    /* print each line, which consists of each column */
    p = head;
    while (p) {
        q = p;
        p = p->down;
        while (q) {
            char pad[6];
            char *tmp = (char *) q;

            if (q->right) {
                memset(pad, '\t', sizeof pad - 1U);
                pad[(sizeof pad) - 1] = 0;
                pad[(colwidth + 7 - strlen(q->line)) / 8] = 0;
            } else {
                pad[0] = '\r';
                pad[1] = '\n';
                pad[2] = 0;
            }
            wrstr(f, q->line);
            wrstr(f, pad);
            q = q->right;
            free(tmp);
        }
    }

    /* reset variables for next time */
    head = tail = NULL;
    colwidth = 0;
    filenames = 0;
}

/* functions to to sort for qsort() */
static int cmp(const void * const a, const void * const b)
{    
    return strcmp(FI_NAME((const FileInfo *) a), FI_NAME((const FileInfo *) b));
}

static int cmp_r(const void * const a, const void * const b)
{    
    return strcmp(FI_NAME((const FileInfo *) b), FI_NAME((const FileInfo *) a));
}

static int cmp_t(const void * const a, const void * const b)
{    
    if (((const FileInfo *) a)->mtime < ((const FileInfo *) b)->mtime) {
        return 1;
    }
    if (((const FileInfo *) a)->mtime > ((const FileInfo *) b)->mtime) {
        return -1;
    }
    return 0;
}

static int cmp_rt(const void * const a, const void * const b)
{    
    return cmp_t(b, a);
}

static int cmp_S(const void * const a, const void * const b)
{    
    if (((const FileInfo *) a)->size < ((const FileInfo *) b)->size) {
        return 1;
    }
    if (((const FileInfo *) a)->size > ((const FileInfo *) b)->size) {
        return -1;
    }
    return 0;
}

static int cmp_rS(const void * const a, const void * const b)
{    
    return cmp_S(b, a);
}

static FileInfo *sreaddir(char **names_pnt)
{
    struct stat st;
    DIR *d;
    struct dirent *de;    
    FileInfo *files_info;
    FileInfo *file_info;
    size_t files_info_size;
    size_t files_info_counter = (size_t) 0U;
    char *names;
    size_t names_size;
    size_t names_counter = (size_t) 0U;
    size_t name_len;
    int (*cmp_func)(const void * const, const void * const);
    
    if ((d = opendir(".")) == NULL) {
        return NULL;
    }
    names_size = CHUNK_SIZE;
    if ((names = malloc(names_size)) == NULL) {
        closedir(d);
        return NULL;
    }
    files_info_size = CHUNK_SIZE / sizeof *files_info;
    if ((files_info = malloc(files_info_size * sizeof *files_info)) == NULL) {
        closedir(d);
        free(names);
        return NULL;
    }
    while ((de = readdir(d)) != NULL) {        
        if (lstat(de->d_name, &st) < 0) {
            continue;
        }
        name_len = strlen(de->d_name) + (size_t) 1U;        
        while (names_counter + name_len >= names_size) {
            char *new_names;
        
            if (name_len >= CHUNK_SIZE) {
                names_size += name_len + CHUNK_SIZE;
            } else {
                names_size += CHUNK_SIZE;
            }
            if ((new_names = realloc(names, names_size)) == NULL) {
                nomem:
                closedir(d);
                free(names);
                free(files_info);
                return NULL;
            }
            names = new_names;
        }
        while ((files_info_counter + (size_t) 1U) >= files_info_size) {
            FileInfo *new_files_info;
            
            files_info_size += (CHUNK_SIZE / sizeof *files_info);
            if ((new_files_info = realloc(files_info, 
                                          files_info_size * sizeof *files_info)) == NULL) {
                goto nomem;
            }
            files_info = new_files_info;
        }        
        memcpy(&names[names_counter], de->d_name, name_len);   /* safe */
        names[names_counter + name_len - 1] = 0;
        file_info = &files_info[files_info_counter];
        file_info->names_pnt = names_pnt;
        file_info->name_offset = names_counter;
        file_info->size = st.st_size;
        file_info->mtime = st.st_mtime;
        file_info->mode = st.st_mode;
        file_info->nlink = st.st_nlink;
        file_info->uid = st.st_uid;        
        file_info->gid = st.st_gid;
        names_counter += name_len;
        files_info_counter++;
    }    
    closedir(d);
    files_info[files_info_counter].name_offset = (size_t) -1;
    *names_pnt = names;    

    if (opt_t) {
        if (opt_r) {
            cmp_func = cmp_rt;
        } else {
            cmp_func = cmp_t;
        }
    } else if (opt_S) {
        if (opt_r) {
            cmp_func = cmp_rS;
        } else {
            cmp_func = cmp_S;
        }
    } else if (opt_r) {
        cmp_func = cmp_r;
    } else {
        cmp_func = cmp;
    }
    qsort(files_info, files_info_counter, sizeof files_info[0], cmp_func);    
    
    return files_info;
}

/* have to change to the directory first ( speed hack for -R ) */
static void listdir(unsigned int depth, int f, const char *name)
{
    FileInfo *dir;
    char *names;
    FileInfo *s;
    FileInfo *r;
    int d;
    
    if (depth >= max_ls_depth || matches >= max_ls_files) {
        return;
    }
    if ((dir = sreaddir(&names)) == NULL) {
        addreply(226, MSG_CANT_READ_FILE, name);
        return;
    }        
    s = dir;
    while (s->name_offset != (size_t) -1) {
        if (FI_NAME(s)[0] != '.') {
            d = listfile(s, NULL);
        } else if (opt_a) {
            d = listfile(s, NULL);
            if (FI_NAME(s)[1] == 0 ||
                (FI_NAME(s)[1] == '.' && FI_NAME(s)[2] == 0)) {
                d = 0;
            }
        } else {
            d = 0;
        }
        if (!d) {
            s->name_offset = (size_t) -1;
        }
        s++;
    }
    outputfiles(f);
    r = dir;
    while (opt_R && r != s) {
        if (r->name_offset != (size_t) -1 && !chdir(FI_NAME(r))) {
            char *alloca_subdir;
            const size_t sizeof_subdir = MAXPATHLEN + 1U;
            
            if ((alloca_subdir = ALLOCA(sizeof_subdir)) == NULL) {
                goto toomany;
            }
            if (SNCHECK(snprintf(alloca_subdir, sizeof_subdir, "%s/%s",
                                 name, FI_NAME(r)), sizeof_subdir)) {
                goto nolist;
            }
#ifndef MINIMAL
            if (modern_listings == 0) {
#endif
#ifdef FANCY_LS_DIRECTORY_HEADERS
                wrstr(f, "\r\n>----------------[");
                wrstr(f, alloca_subdir);
                wrstr(f, "]----------------<\r\n\r\n");
#else
                wrstr(f, "\r\n\r\n");
                wrstr(f, alloca_subdir);
                wrstr(f, ":\r\n\r\n");                    
#endif
#ifndef MINIMAL
            }
#endif
            listdir(depth + 1U, f, alloca_subdir);
            nolist:
            ALLOCA_FREE(alloca_subdir);
            if (matches >= max_ls_files) {
                goto toomany;
            }                
            if (chdir("..")) {    /* defensive in the extreme... */
                chdir(wd);
                if (chdir(name)) {    /* someone rmdir()'d it? */
                    die(421, LOG_ERR, "chdir: %s" ,
                        strerror(errno));
                }
            }
        }
        r++;
    }
    toomany:
    free(names);
    free(dir);
    names = NULL;
}

void donlist(char *arg, int on_ctrl_conn)
{
    int c;

    matches = 0U;

    opt_C = opt_d = opt_F = opt_R = opt_r = opt_t = opt_S = 0;
    if (force_ls_a != 0) {
        opt_a = 1;
    } else {
        opt_a = 0;
    }
    while (isspace((unsigned char)*arg)) {
        arg++;
    }
    while (arg && *arg == '-') {
        while (arg++ && isalnum((unsigned char)*arg)) {
            switch (*arg) {
            case 'a':
                opt_a = 1;
                break;
            case 'l':
                opt_l = 1;
                opt_C = 0;
                break;
            case '1':
                opt_l = opt_C = 0;
                break;
            case 'C':
                opt_l = 0;
                opt_C = 1;
                break;
            case 'F':
                opt_F = 1;
                break;
            case 'R':
                opt_R = 1;
                break;
            case 'd':
                opt_d = 1;
                break;
            case 'r':
                opt_r = 1;
                break;
            case 't':
                opt_t = 1;
                opt_S = 0;
                break;
            case 'S':
                opt_S = 1;
                opt_t = 0;
                break;
            }
        }
        while (isspace((unsigned char)*arg)) {
            arg++;
        }
    }

    if (on_ctrl_conn == 0) {
        if ((c = opendata()) == -1) {
            return;
        }
        doreply();
        if (type == 2) {
            addreply_noformat(0, "ASCII");
        }
    } else {                           /* STAT command */
        c = 1;
        safe_write(c, "213-STAT" CRLF, sizeof "213-STAT" CRLF - 1U);
    }        
    if (arg && *arg) {
        int justone;

        justone = 1;            /* just one argument, so don't print dir name */

        while (arg) {
            glob_t g;
            int a;
            char *endarg = strchr(arg, ' ');

            if (endarg) {
                *endarg++ = 0;
                justone = 0;
            }
#ifdef DEBUG
            if (debug != 0) {
                addreply(226, "Glob: %s", arg);
            }
#endif

            /* Expand ~ here if needed */
            
            alarm(GLOB_TIMEOUT);
            a = sglob(arg,
                      GLOB_BRACE | (opt_a ? GLOB_PERIOD : 0),
                     NULL, &g, max_ls_files + 2, max_ls_depth * 2);
            alarm(0);
            if (a == 0) {
                char **path;

                path = g.gl_pathv;
                if (path && path[0] && path[1]) {
                    justone = 0;
                }
                while (path && *path) {
                    struct stat st;

                    if (stat(*path, &st) == 0) {
                        if (opt_d || !(S_ISDIR(st.st_mode))) {
                            listfile(NULL, *path);
                            **path = 0;
                        }
                    } else {
                        **path = 0;
                    }
                    path++;
                }
                outputfiles(c);    /* in case of opt_C */
                path = g.gl_pathv;
                while (path && *path) {
                    if (matches >= max_ls_files) {
                        break;
                    }
                    if (**path) {
                        if (!justone) {
#ifdef FANCY_LS_DIRECTORY_HEADERS                            
                            wrstr(c, "\r\n>-----------------[");
                            wrstr(c, *path);
                            wrstr(c, "]-----------------<\r\n\r\n");
#else
                            wrstr(c, "\r\n\r\n");
                            wrstr(c, *path);
                            wrstr(c, ":\r\n\r\n");                                            
#endif
                        }
                        if (!chdir(*path)) {
                            listdir(0U, c, *path);
                            chdir(wd);
                        }
                    }
                    path++;
                }
            } else {
                if (a == GLOB_NOSPACE) {
                    addreply(226, MSG_GLOB_NO_MEMORY, arg);
                    addreply_noformat(0, MSG_PROBABLY_DENIED);
                } else if (a == GLOB_ABEND) {
                    addreply(226, MSG_GLOB_READ_ERROR, arg);
                } else if (a != GLOB_NOMATCH) {
                    addreply(226, MSG_GLOB_READ_ERROR, arg);
                    addreply_noformat(0, MSG_PROBABLY_DENIED);                    
                }
            }
            globfree(&g);
            arg = endarg;
        }
    } else {
        if (opt_d) {
            listfile(NULL, ".");
        } else {
            listdir(0U, c, ".");
        }
        outputfiles(c);
    }
    wrstr(c, NULL);
    if (on_ctrl_conn == 0) {
        close(c);
    } else {
        addreply_noformat(213, "End.");
        goto end;
    }
    if (opt_a || opt_C || opt_d || opt_F || opt_l || opt_r || opt_R ||
        opt_t || opt_S)
        addreply(0, "Options: %s%s%s%s%s%s%s%s%s",
                 opt_a ? "-a " : "",
                 opt_C ? "-C " : "",
                 opt_d ? "-d " : "",
                 opt_F ? "-F " : "",
                 opt_l ? "-l " : "",
                 opt_r ? "-r " : "",
                 opt_R ? "-R " : "", opt_S ? "-S " : "",
                 opt_t ? "-t" : "");
    if (matches >= max_ls_files) {
        addreply(226, MSG_LS_TRUNCATED, matches);
    } else {
        addreply(226, MSG_LS_SUCCESS, matches);
    }
    end:
    chdir(wd);
}
